"""Shared defaults for pyannote diarization/separation models."""

PYANNOTE_DIARIZATION_MODEL_ID = "pyannote/speaker-diarization-community-1"
PYANNOTE_DIARIZATION_MODEL_REVISION = "3533c8cf8e369892e6b79ff1bf80f7b0286a54ee"

PYANNOTE_SEPARATION_MODEL_ID = "pyannote/speech-separation-ami-1.0"
PYANNOTE_SEPARATION_MODEL_REVISION = "9486b106945ae0cc0784041a08bfcdba5edadfb9"
